@@ -49,6 +49,7 @@
 #include "src/util.h"
 #include "src/p2putil.h"
 #include "src/mpdu.h"
+#include "src/band.h"
 #include "src/scan.h"
 
 /* User configurable options */
@@ -68,7 +69,6 @@ struct scan_periodic {
 	scan_trigger_func_t trigger;
 	scan_notify_func_t callback;
 	void *userdata;
-	bool retry:1;
 	uint32_t id;
 	bool needs_active_scan:1;
 };
@@ -79,7 +79,21 @@ struct scan_request {
 	scan_notify_func_t callback;
 	void *userdata;
 	scan_destroy_func_t destroy;
+	bool canceled : 1; /* Is scan_cancel being called on this request? */
 	bool passive:1; /* Active or Passive scan? */
+	bool started : 1; /* Has TRIGGER_SCAN succeeded at least once? */
+	bool periodic : 1; /* Started as a periodic scan? */
+	/*
+	 * Set to true if the TRIGGER_SCAN command at the head of the 'cmds'
+	 * queue was acked by the kernel indicating that the scan request was
+	 * successful.  May be set and cleared multiple times during a
+	 * the scan_request lifetime (as each command in the 'cmds' queue is
+	 * issued to the kernel).  Will be false if the current request
+	 * was not started due to an -EBUSY error from the kernel.  Also will
+	 * be false when the scan is complete and GET_SCAN is pending.
+	 */
+	bool triggered : 1;
+	bool in_callback : 1; /* Scan request complete, re-entrancy guard */
 	struct l_queue *cmds;
 	/* The time the current scan was started. Reported in TRIGGER_SCAN */
 	uint64_t start_time_tsf;
@@ -105,18 +119,6 @@ struct scan_context {
 	 * roamed automatically.
 	 */
 	unsigned int get_fw_scan_cmd_id;
-	/*
-	 * Whether the top request in the queue has triggered the current
-	 * scan.  May be set and cleared multiple times during a single
-	 * request.  May be false when the current request is waiting due
-	 * to an EBUSY or an external scan (sr->cmds non-empty), when
-	 * start_cmd_id is non-zero and for a brief moment when GET_SCAN
-	 * is running.
-	 */
-	bool triggered:1;
-	/* Whether any commands from current request's queue have started */
-	bool started:1;
-	bool work_started:1;
 	struct wiphy *wiphy;
 };
 
@@ -163,13 +165,15 @@ static void scan_request_free(struct wiphy_radio_work_item *item)
 static void scan_request_failed(struct scan_context *sc,
 				struct scan_request *sr, int err)
 {
-	l_queue_remove(sc->requests, sr);
+	sr->in_callback = true;
 
 	if (sr->trigger)
 		sr->trigger(err, sr->userdata);
 	else if (sr->callback)
 		sr->callback(err, NULL, NULL, sr->userdata);
 
+	sr->in_callback = false;
+	l_queue_remove(sc->requests, sr);
 	wiphy_radio_work_done(sc->wiphy, sr->work.id);
 }
 
@@ -235,8 +239,6 @@ static void scan_request_triggered(struct l_genl_msg *msg, void *userdata)
 			return;
 		}
 
-		l_queue_remove(sc->requests, sr);
-
 		scan_request_failed(sc, sr, err);
 
 		l_error("Received error during CMD_TRIGGER_SCAN: %s (%d)",
@@ -249,8 +251,8 @@ static void scan_request_triggered(struct l_genl_msg *msg, void *userdata)
 	l_debug("%s scan triggered for wdev %" PRIx64,
 		sr->passive ? "Passive" : "Active", sc->wdev_id);
 
-	sc->triggered = true;
-	sc->started = true;
+	sr->triggered = true;
+	sr->started = true;
 	l_genl_msg_unref(l_queue_pop_head(sr->cmds));
 
 	if (sr->trigger) {
@@ -542,7 +544,7 @@ static void scan_cmds_add(struct l_queue *cmds, struct scan_context *sc,
 	if (params->ssid) {
 		/* direct probe request scan */
 		l_genl_msg_append_attr(cmd, NL80211_ATTR_SSID,
-					strlen(params->ssid), params->ssid);
+					params->ssid_len, params->ssid);
 		l_genl_msg_leave_nested(cmd);
 
 		l_queue_push_tail(cmds, cmd);
@@ -561,6 +563,7 @@ static int scan_request_send_trigger(struct scan_context *sc,
 					struct scan_request *sr)
 {
 	struct l_genl_msg *cmd = l_queue_peek_head(sr->cmds);
+
 	if (!cmd)
 		return -ENOMSG;
 
@@ -583,8 +586,30 @@ static const struct wiphy_radio_work_item_ops work_ops = {
 	.destroy = scan_request_free,
 };
 
+static struct scan_request *scan_request_new(struct scan_context *sc,
+						bool passive,
+						scan_trigger_func_t trigger,
+						scan_notify_func_t notify,
+						void *userdata,
+						scan_destroy_func_t destroy)
+{
+	struct scan_request *sr;
+
+	sr = l_new(struct scan_request, 1);
+	sr->sc = sc;
+	sr->trigger = trigger;
+	sr->callback = notify;
+	sr->userdata = userdata;
+	sr->destroy = destroy;
+	sr->passive = passive;
+	sr->cmds = l_queue_new();
+
+	return sr;
+}
+
 static uint32_t scan_common(uint64_t wdev_id, bool passive,
 				const struct scan_parameters *params,
+				int priority,
 				scan_trigger_func_t trigger,
 				scan_notify_func_t notify, void *userdata,
 				scan_destroy_func_t destroy)
@@ -597,20 +622,14 @@ static uint32_t scan_common(uint64_t wdev_id, bool passive,
 	if (!sc)
 		return 0;
 
-	sr = l_new(struct scan_request, 1);
-	sr->sc = sc;
-	sr->trigger = trigger;
-	sr->callback = notify;
-	sr->userdata = userdata;
-	sr->destroy = destroy;
-	sr->passive = passive;
-	sr->cmds = l_queue_new();
+	sr = scan_request_new(sc, passive, trigger, notify, userdata, destroy);
 
 	scan_cmds_add(sr->cmds, sc, passive, params);
 
 	l_queue_push_tail(sc->requests, sr);
 
-	return wiphy_radio_work_insert(sc->wiphy, &sr->work, 2, &work_ops);
+	return wiphy_radio_work_insert(sc->wiphy, &sr->work,
+					priority, &work_ops);
 }
 
 uint32_t scan_passive(uint64_t wdev_id, struct scan_freq_set *freqs,
@@ -619,8 +638,8 @@ uint32_t scan_passive(uint64_t wdev_id, struct scan_freq_set *freqs,
 {
 	struct scan_parameters params = { .freqs = freqs };
 
-	return scan_common(wdev_id, true, &params, trigger, notify,
-							userdata, destroy);
+	return scan_common(wdev_id, true, &params, WIPHY_WORK_PRIORITY_SCAN,
+				trigger, notify, userdata, destroy);
 }
 
 uint32_t scan_passive_full(uint64_t wdev_id,
@@ -629,8 +648,8 @@ uint32_t scan_passive_full(uint64_t wdev_id,
 			scan_notify_func_t notify, void *userdata,
 			scan_destroy_func_t destroy)
 {
-	return scan_common(wdev_id, true, params, trigger,
-				notify, userdata, destroy);
+	return scan_common(wdev_id, true, params, WIPHY_WORK_PRIORITY_SCAN,
+				trigger, notify, userdata, destroy);
 }
 
 uint32_t scan_active(uint64_t wdev_id, uint8_t *extra_ie, size_t extra_ie_size,
@@ -643,7 +662,7 @@ uint32_t scan_active(uint64_t wdev_id, uint8_t *extra_ie, size_t extra_ie_size,
 	params.extra_ie = extra_ie;
 	params.extra_ie_size = extra_ie_size;
 
-	return scan_common(wdev_id, false, &params,
+	return scan_common(wdev_id, false, &params, WIPHY_WORK_PRIORITY_SCAN,
 					trigger, notify, userdata, destroy);
 }
 
@@ -652,10 +671,138 @@ uint32_t scan_active_full(uint64_t wdev_id,
 			scan_trigger_func_t trigger, scan_notify_func_t notify,
 			void *userdata, scan_destroy_func_t destroy)
 {
-	return scan_common(wdev_id, false, params,
+	return scan_common(wdev_id, false, params, WIPHY_WORK_PRIORITY_SCAN,
 					trigger, notify, userdata, destroy);
 }
 
+static void scan_add_owe_freq(struct scan_freq_set *freqs,
+				const struct scan_bss *bss)
+{
+	int freq;
+
+	if (bss->owe_trans->oper_class)
+		freq = oci_to_frequency(bss->owe_trans->oper_class,
+					bss->owe_trans->channel);
+	else
+		freq = bss->frequency;
+
+	L_WARN_ON(freq < 0);
+
+	scan_freq_set_add(freqs, freq);
+}
+
+static void add_owe_scan_cmd(struct scan_context *sc, struct scan_request *sr,
+				bool ignore_flush,
+				struct scan_freq_set *freqs,
+				const struct scan_bss *bss)
+{
+	struct l_genl_msg *cmd;
+	struct scan_parameters params = {};
+	struct scan_freq_set *tmp;
+
+	if (!freqs) {
+		tmp = scan_freq_set_new();
+
+		scan_add_owe_freq(tmp, bss);
+
+		params.freqs = tmp;
+	} else
+		params.freqs = freqs;
+
+	params.ssid = bss->owe_trans->ssid;
+	params.ssid_len = bss->owe_trans->ssid_len;
+	params.flush = true;
+
+	cmd = scan_build_cmd(sc, ignore_flush, false, &params);
+
+	l_genl_msg_enter_nested(cmd, NL80211_ATTR_SCAN_SSIDS);
+	l_genl_msg_append_attr(cmd, 0, params.ssid_len, params.ssid);
+	l_genl_msg_leave_nested(cmd);
+
+	l_queue_push_tail(sr->cmds, cmd);
+
+	if (!freqs)
+		scan_freq_set_free(tmp);
+}
+
+uint32_t scan_owe_hidden(uint64_t wdev_id, struct l_queue *list,
+			scan_trigger_func_t trigger, scan_notify_func_t notify,
+			void *userdata, scan_destroy_func_t destroy)
+{
+	struct scan_context *sc;
+	struct scan_request *sr;
+	struct scan_freq_set *freqs;
+	const struct l_queue_entry *entry;
+	const uint8_t *ssid = NULL;
+	size_t ssid_len;
+	bool same_ssid = true;
+	struct scan_bss *bss;
+	bool ignore_flush = false;
+
+	sc = l_queue_find(scan_contexts, scan_context_match, &wdev_id);
+
+	if (!sc)
+		return 0;
+
+	sr = scan_request_new(sc, false, trigger, notify, userdata, destroy);
+
+	freqs = scan_freq_set_new();
+
+	/*
+	 * Start building up a frequency list if all SSIDs are the same. This
+	 * is hopefully the common case and will allow a single scan command.
+	 */
+	for (entry = l_queue_get_entries(list); entry; entry = entry->next) {
+		bss = entry->data;
+
+		scan_add_owe_freq(freqs, bss);
+
+		/* First */
+		if (!ssid) {
+			ssid = bss->owe_trans->ssid;
+			ssid_len = bss->owe_trans->ssid_len;
+			continue;
+		}
+
+		if (ssid_len == bss->owe_trans->ssid_len &&
+				!memcmp(ssid, bss->owe_trans->ssid,
+				bss->owe_trans->ssid_len))
+			continue;
+
+		same_ssid = false;
+		break;
+	}
+
+	if (same_ssid) {
+		bss = l_queue_peek_head(list);
+
+		add_owe_scan_cmd(sc, sr, ignore_flush, freqs, bss);
+
+		scan_freq_set_free(freqs);
+
+		goto done;
+	}
+
+	scan_freq_set_free(freqs);
+
+	/* SSIDs differed, use separate scan commands. */
+	for (entry = l_queue_get_entries(list); entry; entry = entry->next) {
+		bss = entry->data;
+
+		add_owe_scan_cmd(sc, sr, ignore_flush, NULL, bss);
+
+		/* Ignore flush on all subsequent commands */
+		if (!ignore_flush)
+			ignore_flush = true;
+	}
+
+done:
+	l_queue_push_tail(sc->requests, sr);
+
+	return wiphy_radio_work_insert(sc->wiphy, &sr->work,
+					WIPHY_WORK_PRIORITY_SCAN, &work_ops);
+}
+
 bool scan_cancel(uint64_t wdev_id, uint32_t id)
 {
 	struct scan_context *sc;
@@ -671,23 +818,33 @@ bool scan_cancel(uint64_t wdev_id, uint32_t id)
 	if (!sr)
 		return false;
 
+	/* We're in the callback and about to be removed, invoke destroy now */
+	if (sr->in_callback)
+		goto call_destroy;
+
 	/* If already triggered, just zero out the callback */
-	if (sr == l_queue_peek_head(sc->requests) && sc->triggered) {
-		l_debug("Scan is at the top of the queue and triggered");
+	if (sr->triggered) {
+		l_debug("Scan has been triggered, wait for it to complete");
 
 		sr->callback = NULL;
-
-		if (sr->destroy) {
-			sr->destroy(sr->userdata);
-			sr->destroy = NULL;
-		}
-
-		return true;
+		goto call_destroy;
 	}
 
-	/* If we already sent the trigger command, cancel the scan */
-	if (sr == l_queue_peek_head(sc->requests)) {
-		l_debug("Scan is at the top of the queue, but not triggered");
+	/*
+	 * Takes care of the following cases:
+	 * 1. If TRIGGER_SCAN is in flight
+	 * 2. TRIGGER_SCAN sent but bounced with -EBUSY
+	 * 3. Scan request is done but GET_SCAN is still pending
+	 *
+	 * For case 3, we can easily cancel the command and proceed with the
+	 * other pending requests.  For case 1 & 2, the subsequent pending
+	 * request might bounce off with an -EBUSY.
+	 */
+	if (wiphy_radio_work_is_running(sc->wiphy, sr->work.id)) {
+		l_debug("Scan is already started");
+
+		/* l_genl_family_cancel will trigger destroy callbacks */
+		sr->canceled = true;
 
 		if (sc->start_cmd_id)
 			l_genl_family_cancel(nl80211, sc->start_cmd_id);
@@ -696,14 +853,20 @@ bool scan_cancel(uint64_t wdev_id, uint32_t id)
 			l_genl_family_cancel(nl80211, sc->get_scan_cmd_id);
 
 		sc->start_cmd_id = 0;
-		l_queue_remove(sc->requests, sr);
-		sc->started = false;
-		sc->work_started = false;
-	} else
-		l_queue_remove(sc->requests, sr);
+		sc->get_scan_cmd_id = 0;
+	}
 
+	l_queue_remove(sc->requests, sr);
 	wiphy_radio_work_done(sc->wiphy, sr->work.id);
 
+	return true;
+
+call_destroy:
+	if (sr->destroy) {
+		sr->destroy(sr->userdata);
+		sr->destroy = NULL;
+	}
+
 	return true;
 }
 
@@ -736,29 +899,41 @@ static bool scan_periodic_notify(int err, struct l_queue *bss_list,
 	return false;
 }
 
+static void scan_periodic_destroy(void *user_data)
+{
+	struct scan_context *sc = user_data;
+
+	sc->sp.id = 0;
+}
+
 static bool scan_periodic_queue(struct scan_context *sc)
 {
-	if (!l_queue_isempty(sc->requests)) {
-		sc->sp.retry = true;
-		return false;
-	}
+	struct scan_parameters params = {};
+	struct scan_request *sr;
 
 	if (sc->sp.needs_active_scan && known_networks_has_hidden()) {
-		struct scan_parameters params = {
-			.randomize_mac_addr_hint = true
-		};
+		params.randomize_mac_addr_hint = true;
 
 		sc->sp.needs_active_scan = false;
-
-		sc->sp.id = scan_active_full(sc->wdev_id, &params,
-						scan_periodic_triggered,
-						scan_periodic_notify, sc, NULL);
+		sc->sp.id = scan_common(sc->wdev_id, false, &params,
+					WIPHY_WORK_PRIORITY_PERIODIC_SCAN,
+					scan_periodic_triggered,
+					scan_periodic_notify, sc,
+					scan_periodic_destroy);
 	} else
-		sc->sp.id = scan_passive(sc->wdev_id, NULL,
-						scan_periodic_triggered,
-						scan_periodic_notify, sc, NULL);
+		sc->sp.id = scan_common(sc->wdev_id, true, &params,
+					WIPHY_WORK_PRIORITY_PERIODIC_SCAN,
+					scan_periodic_triggered,
+					scan_periodic_notify, sc,
+					scan_periodic_destroy);
 
-	return sc->sp.id != 0;
+	if (!sc->sp.id)
+		return false;
+
+	sr = l_queue_peek_tail(sc->requests);
+	sr->periodic = true;
+
+	return true;
 }
 
 static bool scan_periodic_is_disabled(void)
@@ -784,7 +959,7 @@ void scan_periodic_start(uint64_t wdev_id, scan_trigger_func_t trigger,
 	sc = l_queue_find(scan_contexts, scan_context_match, &wdev_id);
 
 	if (!sc) {
-		l_error("scan_periodic_start called without scan_wdev_add");
+		l_error("%s called without scan_wdev_add", __func__);
 		return;
 	}
 
@@ -828,7 +1003,6 @@ bool scan_periodic_stop(uint64_t wdev_id)
 	sc->sp.trigger = NULL;
 	sc->sp.callback = NULL;
 	sc->sp.userdata = NULL;
-	sc->sp.retry = false;
 	sc->sp.needs_active_scan = false;
 
 	return true;
@@ -843,11 +1017,8 @@ uint64_t scan_get_triggered_time(uint64_t wdev_id, uint32_t id)
 	if (!sc)
 		return 0;
 
-	if (!sc->triggered)
-		return 0;
-
 	sr = l_queue_find(sc->requests, scan_request_match, L_UINT_TO_PTR(id));
-	if (!sr)
+	if (!sr || !sr->triggered)
 		return 0;
 
 	return sr->start_time_tsf;
@@ -857,7 +1028,17 @@ static void scan_periodic_timeout(struct l_timeout *timeout, void *user_data)
 {
 	struct scan_context *sc = user_data;
 
-	l_debug("scan_periodic_timeout: %" PRIx64, sc->wdev_id);
+	l_debug("%" PRIx64, sc->wdev_id);
+
+	/*
+	 * Timeout triggered before periodic scan could even start, just rearm
+	 * with the same interval.
+	 */
+	if (sc->sp.id) {
+		l_debug("Periodic scan timer called before scan could start!");
+		scan_periodic_rearm(sc);
+		return;
+	}
 
 	sc->sp.interval *= 2;
 	if (sc->sp.interval > SCAN_MAX_INTERVAL)
@@ -891,43 +1072,72 @@ static bool start_next_scan_request(struct wiphy_radio_work_item *item)
 						struct scan_request, work);
 	struct scan_context *sc = sr->sc;
 
-	sc->work_started = true;
-
 	if (sc->state != SCAN_STATE_NOT_RUNNING)
 		return false;
 
 	if (!scan_request_send_trigger(sc, sr))
 		return false;
 
-	sc->work_started = false;
-
 	scan_request_failed(sc, sr, -EIO);
 
-	if (sc->sp.retry) {
-		sc->sp.retry = false;
-		scan_periodic_queue(sc);
-	}
-
 	return true;
 }
 
-static bool scan_parse_vendor_specific(struct scan_bss *bss, const void *data,
+static void scan_parse_vendor_specific(struct scan_bss *bss, const void *data,
 					uint16_t len)
 {
-	if (!bss->wpa && is_ie_wpa_ie(data, len))
+	uint16_t cost_level;
+	uint16_t cost_flags;
+	bool dgaf_disable;
+
+	if (!bss->wpa && is_ie_wpa_ie(data, len)) {
 		bss->wpa = l_memdup(data - 2, len + 2);
-	else if (!bss->osen && is_ie_wfa_ie(data, len, IE_WFA_OI_OSEN))
+		return;
+	}
+
+	if (!bss->osen && is_ie_wfa_ie(data, len, IE_WFA_OI_OSEN)) {
 		bss->osen = l_memdup(data - 2, len + 2);
-	else if (is_ie_wfa_ie(data, len, IE_WFA_OI_HS20_INDICATION)) {
+		return;
+	}
+
+	if (is_ie_wfa_ie(data, len, IE_WFA_OI_HS20_INDICATION)) {
 		if (ie_parse_hs20_indication_from_data(data - 2, len + 2,
-					&bss->hs20_version, NULL, NULL) < 0)
-			return false;
+					&bss->hs20_version, NULL, NULL,
+					&dgaf_disable) < 0)
+			return;
 
+		bss->hs20_dgaf_disable = dgaf_disable;
 		bss->hs20_capable = true;
-	} else
-		return false;
+		return;
+	}
 
-	return true;
+	if (is_ie_wfa_ie(data, len, IE_WFA_OI_OWE_TRANSITION)) {
+		_auto_(l_free) struct ie_owe_transition_info *owe_trans =
+				l_new(struct ie_owe_transition_info, 1);
+
+		if (ie_parse_owe_transition(data - 2, len + 2, owe_trans) < 0)
+			return;
+
+		if (owe_trans->oper_class &&
+				oci_to_frequency(owe_trans->oper_class,
+						owe_trans->channel) < 0)
+			return;
+
+		bss->owe_trans = l_steal_ptr(owe_trans);
+		return;
+	}
+
+	if (is_ie_wfa_ie(data, len, IE_WFA_OI_CONFIGURATOR_CONNECTIVITY))
+		bss->dpp_configurator = true;
+
+	if (!ie_parse_network_cost(data, len, &cost_level, &cost_flags)) {
+		bss->cost_level = cost_level;
+		bss->cost_flags = cost_flags;
+		return;
+	}
+
+	if (is_ie_default_sae_group_oui(data, len))
+		bss->force_default_sae_group = true;
 }
 
 /*
@@ -1019,7 +1229,6 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 
 			break;
 		case IE_TYPE_VENDOR_SPECIFIC:
-			/* Interested only in WPA/WFA IE from Vendor data */
 			scan_parse_vendor_specific(bss, iter.data, iter.len);
 			break;
 		case IE_TYPE_MOBILITY_DOMAIN:
@@ -1079,6 +1288,23 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 			bss->rc_ie = l_memdup(iter.data - 2, iter.len + 2);
 
 			break;
+
+		case IE_TYPE_EXTENDED_CAPABILITIES:
+			/* 802.11-2020 9.4.2.26
+			 *
+			 * "The length of the Extended Capabilities field is
+			 * variable. If fewer bits are received in an Extended
+			 * Capabilities field than shown in Table 9-153, the
+			 * rest of the Extended Capabilities field bits are
+			 * assumed to be zero"
+			 *
+			 * Currently only Proxy ARP bit (12) is checked, and if
+			 * not found, this is not a fatal error.
+			 */
+			if (iter.len < 2)
+				break;
+
+			bss->proxy_arp = test_bit(iter.data, 12);
 		}
 	}
 
@@ -1366,6 +1592,7 @@ void scan_bss_free(struct scan_bss *bss)
 	l_free(bss->osen);
 	l_free(bss->rc_ie);
 	l_free(bss->wfd);
+	l_free(bss->owe_trans);
 
 	switch (bss->source_frame) {
 	case SCAN_BSS_PROBE_RESP:
@@ -1434,6 +1661,10 @@ int scan_bss_rank_compare(const void *a, const void *b, void *user_data)
 {
 	const struct scan_bss *new_bss = a, *bss = b;
 
+	if (bss->rank == new_bss->rank)
+		return (bss->signal_strength >
+					new_bss->signal_strength) ? 1 : -1;
+
 	return (bss->rank > new_bss->rank) ? 1 : -1;
 }
 
@@ -1490,34 +1721,34 @@ static void scan_finished(struct scan_context *sc,
 				struct scan_request *sr)
 {
 	bool new_owner = false;
+	scan_notify_func_t callback = sr ? sr->callback : sc->sp.callback;
+	void *userdata = sr ? sr->userdata : sc->sp.userdata;
 
 	if (bss_list)
 		discover_hidden_network_bsses(sc, bss_list);
 
-	if  (sr) {
-		l_queue_remove(sc->requests, sr);
-		sc->started = false;
-		sc->work_started = false;
+	if (sr)
+		sr->in_callback = true;
 
-		if (sr->callback)
-			new_owner = sr->callback(err, bss_list,
-							freqs, sr->userdata);
-
-		/*
-		 * Can start a new scan now that we've removed this one from
-		 * the queue.  If this were an external scan request (sr NULL)
-		 * then the SCAN_FINISHED or SCAN_ABORTED handler would have
-		 * taken care of sending the next command for a new or ongoing
-		 * scan, or scheduling the next periodic scan.
-		 */
-		wiphy_radio_work_done(sc->wiphy, sr->work.id);
-	} else if (sc->sp.callback)
-		new_owner = sc->sp.callback(err, bss_list,
-						freqs, sc->sp.userdata);
+	if (callback)
+		new_owner = callback(err, bss_list, freqs, userdata);
 
 	if (bss_list && !new_owner)
 		l_queue_destroy(bss_list,
 				(l_queue_destroy_func_t) scan_bss_free);
+
+	if (!sr)
+		return;
+
+	/*
+	 * Can start a new scan now that we've removed this one from the
+	 * queue.  If this were an external scan request (sr NULL) then the
+	 * SCAN_FINISHED or SCAN_ABORTED handler would have taken care of
+	 * sending the next command for a new or ongoing scan.
+	 */
+	sr->in_callback = false;
+	l_queue_remove(sc->requests, sr);
+	wiphy_radio_work_done(sc->wiphy, sr->work.id);
 }
 
 static void get_scan_done(void *user)
@@ -1529,7 +1760,7 @@ static void get_scan_done(void *user)
 
 	sc->get_scan_cmd_id = 0;
 
-	if (l_queue_peek_head(sc->requests) == results->sr)
+	if (!results->sr || !results->sr->canceled)
 		scan_finished(sc, 0, results->bss_list,
 						results->freqs, results->sr);
 	else
@@ -1598,8 +1829,6 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 
 	cmd = l_genl_msg_get_command(msg);
 
-	l_debug("Scan notification %s(%u)", nl80211cmd_to_string(cmd), cmd);
-
 	if (nl80211_parse_attrs(msg, NL80211_ATTR_WDEV, &wdev_id,
 					NL80211_ATTR_WIPHY, &wiphy_id,
 					NL80211_ATTR_UNSPEC) < 0)
@@ -1609,6 +1838,8 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 	if (!sc)
 		return;
 
+	l_debug("Scan notification %s(%u)", nl80211cmd_to_string(cmd), cmd);
+
 	if (!l_genl_attr_init(&attr, msg))
 		return;
 
@@ -1634,16 +1865,14 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 		struct l_genl_msg *scan_msg;
 		struct scan_results *results;
 		bool send_next = false;
+		bool retry = false;
 		bool get_results = false;
 
-		if (sc->state == SCAN_STATE_NOT_RUNNING)
-			break;
-
 		sc->state = SCAN_STATE_NOT_RUNNING;
 
 		/* Was this our own scan or an external scan */
-		if (sc->triggered) {
-			sc->triggered = false;
+		if (sr && sr->triggered) {
+			sr->triggered = false;
 
 			if (!sr->callback) {
 				scan_finished(sc, -ECANCELED, NULL, NULL, sr);
@@ -1667,28 +1896,31 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 			if (sc->sp.callback)
 				get_results = true;
 
-			/* An external scan may have flushed our results */
-			if (sc->started && scan_parse_flush_flag_from_msg(msg))
+			/*
+			 * Drop the ongoing scan if an external scan flushed
+			 * our results.  Otherwise, try to retry the trigger
+			 * request if it failed with an -EBUSY.
+			 */
+			if (sr && sr->started &&
+					scan_parse_flush_flag_from_msg(msg))
 				scan_finished(sc, -EAGAIN, NULL, NULL, sr);
 			else
-				send_next = true;
+				retry = true;
 
 			sr = NULL;
 		}
 
 		/*
-		 * Send the next command of an ongoing request, or continue with
-		 * a previously busy scan attempt due to an external scan. A
-		 * temporary scan_request object is used (rather than 'sr')
-		 * because the state of 'sr' tells us if the results should
-		 * be used either as normal scan results, or to take advantage
-		 * of the external scan as a 'free' periodic scan of sorts.
+		 * Send the next command of an ongoing request, or continue
+		 * with a previously busy scan attempt due to an external
+		 * scan.
 		 */
-		if (sc->work_started && send_next) {
+		if (send_next || retry) {
 			struct scan_request *next = l_queue_peek_head(
 								sc->requests);
 
-			if (next)
+			if (next && wiphy_radio_work_is_running(sc->wiphy,
+								next->work.id))
 				start_next_scan_request(&next->work);
 		}
 
@@ -1725,16 +1957,25 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 		break;
 
 	case NL80211_CMD_SCAN_ABORTED:
-		if (sc->state == SCAN_STATE_NOT_RUNNING)
-			break;
-
 		sc->state = SCAN_STATE_NOT_RUNNING;
 
-		if (sc->triggered) {
-			sc->triggered = false;
+		/*
+		 * If there's nothing pending, then most likely an external
+		 * scan got aborted.  We don't care, ignore.
+		 */
+		if (!sr)
+			break;
 
-			scan_finished(sc, -ECANCELED, NULL, NULL, sr);
-		} else {
+		if (sr->triggered) {
+			sr->triggered = false;
+
+			/* If periodic scan, don't report the abort */
+			if (sr->periodic)
+				wiphy_radio_work_done(sc->wiphy, sr->work.id);
+			else
+				scan_finished(sc, -ECANCELED, NULL, NULL, sr);
+		} else if (wiphy_radio_work_is_running(sc->wiphy,
+							sr->work.id)) {
 			/*
 			 * If this was an external scan that got aborted
 			 * we may be able to now queue our own scan although
@@ -1742,8 +1983,7 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 			 * hardware or the driver because of another activity
 			 * starting in which case we should just get an EBUSY.
 			 */
-			if (sc->work_started)
-				start_next_scan_request(&sr->work);
+			start_next_scan_request(&sr->work);
 		}
 
 		break;
@@ -1816,359 +2056,6 @@ bool scan_get_firmware_scan(uint64_t wdev_id, scan_notify_func_t notify,
 	return true;
 }
 
-uint8_t scan_freq_to_channel(uint32_t freq, enum scan_band *out_band)
-{
-	uint32_t channel = 0;
-
-	if (freq >= 2412 && freq <= 2484) {
-		if (freq == 2484)
-			channel = 14;
-		else {
-			channel = freq - 2407;
-
-			if (channel % 5)
-				return 0;
-
-			channel /= 5;
-		}
-
-		if (out_band)
-			*out_band = SCAN_BAND_2_4_GHZ;
-
-		return channel;
-	}
-
-	if (freq >= 5005 && freq < 5900) {
-		if (freq % 5)
-			return 0;
-
-		channel = (freq - 5000) / 5;
-
-		if (out_band)
-			*out_band = SCAN_BAND_5_GHZ;
-
-		return channel;
-	}
-
-	if (freq >= 4905 && freq < 5000) {
-		if (freq % 5)
-			return 0;
-
-		channel = (freq - 4000) / 5;
-
-		if (out_band)
-			*out_band = SCAN_BAND_5_GHZ;
-
-		return channel;
-	}
-
-	return 0;
-}
-
-uint32_t scan_channel_to_freq(uint8_t channel, enum scan_band band)
-{
-	if (band == SCAN_BAND_2_4_GHZ) {
-		if (channel >= 1 && channel <= 13)
-			return 2407 + 5 * channel;
-
-		if (channel == 14)
-			return 2484;
-	}
-
-	if (band == SCAN_BAND_5_GHZ) {
-		if (channel >= 1 && channel <= 179)
-			return 5000 + 5 * channel;
-
-		if (channel >= 181 && channel <= 199)
-			return 4000 + 5 * channel;
-	}
-
-	return 0;
-}
-
-static const char *const oper_class_us_codes[] = {
-	"US", "CA"
-};
-
-static const char *const oper_class_eu_codes[] = {
-	"AL", "AM", "AT", "AZ", "BA", "BE", "BG", "BY", "CH", "CY", "CZ", "DE",
-	"DK", "EE", "EL", "ES", "FI", "FR", "GE", "HR", "HU", "IE", "IS", "IT",
-	"LI", "LT", "LU", "LV", "MD", "ME", "MK", "MT", "NL", "NO", "PL", "PT",
-	"RO", "RS", "RU", "SE", "SI", "SK", "TR", "UA", "UK"
-};
-
-/* Annex E, table E-1 */
-static const uint8_t oper_class_us_to_global[] = {
-	[1]  = 115, [2]  = 118, [3]  = 124, [4]  = 121,
-	[5]  = 125, [6]  = 103, [7]  = 103, [8]  = 102,
-	[9]  = 102, [10] = 101, [11] = 101, [12] = 81,
-	[13] = 94,  [14] = 95,  [15] = 96,  [22] = 116,
-	[23] = 119, [24] = 122, [25] = 126, [26] = 126,
-	[27] = 117, [28] = 120, [29] = 123, [30] = 127,
-	[31] = 127, [32] = 83,  [33] = 84,  [34] = 180,
-	/* 128 - 130 is a 1 to 1 mapping */
-};
-
-/* Annex E, table E-2 */
-static const uint8_t oper_class_eu_to_global[] = {
-	[1]  = 115, [2]  = 118, [3]  = 121, [4]  = 81,
-	[5]  = 116, [6]  = 119, [7]  = 122, [8]  = 117,
-	[9]  = 120, [10] = 123, [11] = 83,  [12] = 84,
-	[17] = 125, [18] = 130,
-	/* 128 - 130 is a 1 to 1 mapping */
-};
-
-/* Annex E, table E-3 */
-static const uint8_t oper_class_jp_to_global[] = {
-	[1]  = 115, [2]  = 112, [3]  = 112, [4]  = 112,
-	[5]  = 112, [6]  = 112, [7]  = 109, [8]  = 109,
-	[9]  = 109, [10] = 109, [11] = 109, [12] = 113,
-	[13] = 113, [14] = 113, [15] = 113, [16] = 110,
-	[17] = 110, [18] = 110, [19] = 110, [20] = 110,
-	[21] = 114, [22] = 114, [23] = 114, [24] = 114,
-	[25] = 111, [26] = 111, [27] = 111, [28] = 111,
-	[29] = 111, [30] = 81,  [31] = 82,  [32] = 118,
-	[33] = 118, [34] = 121, [35] = 121, [36] = 116,
-	[37] = 119, [38] = 119, [39] = 122, [40] = 122,
-	[41] = 117, [42] = 120, [43] = 120, [44] = 123,
-	[45] = 123, [46] = 104, [47] = 104, [48] = 104,
-	[49] = 104, [50] = 104, [51] = 105, [52] = 105,
-	[53] = 105, [54] = 105, [55] = 105, [56] = 83,
-	[57] = 84,  [58] = 121, [59] = 180,
-	/* 128 - 130 is a 1 to 1 mapping */
-};
-
-/* Annex E, table E-4 (only 2.4GHz and 4.9 / 5GHz bands) */
-static const enum scan_band oper_class_to_band_global[] = {
-	[81 ... 84]   = SCAN_BAND_2_4_GHZ,
-	[104 ... 130] = SCAN_BAND_5_GHZ,
-};
-
-/* Annex E, table E-5 */
-static const uint8_t oper_class_cn_to_global[] = {
-	[1]  = 115, [2]  = 118, [3]  = 125, [4]  = 116,
-	[5]  = 119, [6]  = 126, [7]  = 81,  [8]  = 83,
-	[9]  = 84,
-	/* 128 - 130 is a 1 to 1 mapping */
-};
-
-enum scan_band scan_oper_class_to_band(const uint8_t *country,
-					uint8_t oper_class)
-{
-	unsigned int i;
-	int table = 0;
-
-	if (country && country[2] >= 1 && country[2] <= 5)
-		table = country[2];
-	else if (country) {
-		for (i = 0; i < L_ARRAY_SIZE(oper_class_us_codes); i++)
-			if (!memcmp(oper_class_us_codes[i], country, 2)) {
-				/* Use table E-1 */
-				table = 1;
-				break;
-			}
-
-		for (i = 0; i < L_ARRAY_SIZE(oper_class_eu_codes); i++)
-			if (!memcmp(oper_class_eu_codes[i], country, 2)) {
-				/* Use table E-2 */
-				table = 2;
-				break;
-			}
-
-		if (!memcmp("JP", country, 2))
-			/* Use table E-3 */
-			table = 3;
-
-		if (!memcmp("CN", country, 2))
-			/* Use table E-5 */
-			table = 5;
-	}
-
-	switch (table) {
-	case 1:
-		if (oper_class < L_ARRAY_SIZE(oper_class_us_to_global))
-			oper_class = oper_class_us_to_global[oper_class];
-		break;
-	case 2:
-		if (oper_class < L_ARRAY_SIZE(oper_class_eu_to_global))
-			oper_class = oper_class_eu_to_global[oper_class];
-		break;
-	case 3:
-		if (oper_class < L_ARRAY_SIZE(oper_class_jp_to_global))
-			oper_class = oper_class_jp_to_global[oper_class];
-		break;
-	case 5:
-		if (oper_class < L_ARRAY_SIZE(oper_class_cn_to_global))
-			oper_class = oper_class_cn_to_global[oper_class];
-		break;
-	}
-
-	if (oper_class < L_ARRAY_SIZE(oper_class_to_band_global))
-		return oper_class_to_band_global[oper_class];
-	else
-		return 0;
-}
-
-struct scan_freq_set {
-	uint16_t channels_2ghz;
-	struct l_uintset *channels_5ghz;
-};
-
-struct scan_freq_set *scan_freq_set_new(void)
-{
-	struct scan_freq_set *ret = l_new(struct scan_freq_set, 1);
-
-	/* 802.11-2012, 8.4.2.10 hints that 200 is the largest channel number */
-	ret->channels_5ghz = l_uintset_new_from_range(1, 200);
-
-	return ret;
-}
-
-void scan_freq_set_free(struct scan_freq_set *freqs)
-{
-	l_uintset_free(freqs->channels_5ghz);
-	l_free(freqs);
-}
-
-bool scan_freq_set_add(struct scan_freq_set *freqs, uint32_t freq)
-{
-	enum scan_band band;
-	uint8_t channel;
-
-	channel = scan_freq_to_channel(freq, &band);
-	if (!channel)
-		return false;
-
-	switch (band) {
-	case SCAN_BAND_2_4_GHZ:
-		freqs->channels_2ghz |= 1 << (channel - 1);
-		return true;
-	case SCAN_BAND_5_GHZ:
-		return l_uintset_put(freqs->channels_5ghz, channel);
-	}
-
-	return false;
-}
-
-bool scan_freq_set_contains(const struct scan_freq_set *freqs, uint32_t freq)
-{
-	enum scan_band band;
-	uint8_t channel;
-
-	channel = scan_freq_to_channel(freq, &band);
-	if (!channel)
-		return false;
-
-	switch (band) {
-	case SCAN_BAND_2_4_GHZ:
-		return freqs->channels_2ghz & (1 << (channel - 1));
-	case SCAN_BAND_5_GHZ:
-		return l_uintset_contains(freqs->channels_5ghz, channel);
-	}
-
-	return false;
-}
-
-uint32_t scan_freq_set_get_bands(struct scan_freq_set *freqs)
-{
-	uint32_t bands = 0;
-	uint32_t max;
-
-	if (freqs->channels_2ghz)
-		bands |= SCAN_BAND_2_4_GHZ;
-
-	max = l_uintset_get_max(freqs->channels_5ghz);
-
-	if (l_uintset_find_min(freqs->channels_5ghz) <= max)
-		bands |= SCAN_BAND_5_GHZ;
-
-	return bands;
-}
-
-static void scan_channels_5ghz_add(uint32_t channel, void *user_data)
-{
-	struct l_uintset *to = user_data;
-
-	l_uintset_put(to, channel);
-}
-
-void scan_freq_set_merge(struct scan_freq_set *to,
-					const struct scan_freq_set *from)
-{
-	to->channels_2ghz |= from->channels_2ghz;
-
-	l_uintset_foreach(from->channels_5ghz, scan_channels_5ghz_add,
-							to->channels_5ghz);
-}
-
-bool scan_freq_set_isempty(const struct scan_freq_set *set)
-{
-	if (set->channels_2ghz == 0 && l_uintset_isempty(set->channels_5ghz))
-		return true;
-
-	return false;
-}
-
-struct channels_5ghz_foreach_data {
-	scan_freq_set_func_t func;
-	void *user_data;
-};
-
-static void scan_channels_5ghz_frequency(uint32_t channel, void *user_data)
-{
-	const struct channels_5ghz_foreach_data *channels_5ghz_data = user_data;
-	uint32_t freq;
-
-	freq = scan_channel_to_freq(channel, SCAN_BAND_5_GHZ);
-
-	channels_5ghz_data->func(freq, channels_5ghz_data->user_data);
-}
-
-void scan_freq_set_foreach(const struct scan_freq_set *freqs,
-				scan_freq_set_func_t func, void *user_data)
-{
-	struct channels_5ghz_foreach_data data = { };
-	uint8_t channel;
-	uint32_t freq;
-
-	if (unlikely(!freqs || !func))
-		return;
-
-	data.func = func;
-	data.user_data = user_data;
-
-	l_uintset_foreach(freqs->channels_5ghz, scan_channels_5ghz_frequency,
-									&data);
-
-	if (!freqs->channels_2ghz)
-		return;
-
-	for (channel = 1; channel <= 14; channel++) {
-		if (freqs->channels_2ghz & (1 << (channel - 1))) {
-			freq = scan_channel_to_freq(channel, SCAN_BAND_2_4_GHZ);
-
-			func(freq, user_data);
-		}
-	}
-}
-
-void scan_freq_set_constrain(struct scan_freq_set *set,
-					const struct scan_freq_set *constraint)
-{
-	struct l_uintset *intersection;
-
-	intersection = l_uintset_intersect(constraint->channels_5ghz,
-							set->channels_5ghz);
-	if (!intersection)
-		/* This shouldn't ever be the case. */
-		return;
-
-	l_uintset_free(set->channels_5ghz);
-	set->channels_5ghz = intersection;
-
-	set->channels_2ghz &= constraint->channels_2ghz;
-}
-
 bool scan_wdev_add(uint64_t wdev_id)
 {
 	struct scan_context *sc;
@@ -2239,7 +2126,7 @@ static int scan_init(void)
 	return 0;
 }
 
-static void scan_exit()
+static void scan_exit(void)
 {
 	l_queue_destroy(scan_contexts,
 				(l_queue_destroy_func_t) scan_context_free);
